# $Id: S5DAGGeneratorA.py,v 1.21 2005/10/12 18:41:36 kipp Exp $

#
# Preamble.
#

from glue import lal
from glue import segments
import onasys.DAGGenerator
from onasys import onasysd
import logging
from math import ceil, floor
import os
import popen2


#
# Helpful functions.
#

def compute_boundaries(config, sciseg, analseg):
	"""
	Given configuration information, a science segment, and an unused
	analysis segment, compute the corrected boundaries of the analysis
	segment and the surrounding data segment.
	"""
	# Initialize the segment length discretizers
	analsteps = onasysd.StepSegment(config.analysis_length_minimum, config.analysis_length_increment)
	ltranssteps = onasysd.StepSegment(config.transient_left_minimum, config.transient_left_increment)
	rtranssteps = onasysd.StepSegment(config.transient_right_minimum, config.transient_right_increment)

	# Move the analysis segment boundaries away from the edges of the
	# science segment to make room for the minimum allowed transients,
	# and truncate the result to an allowed size.
	if sciseg.duration() <= config.transient_left_minimum + config.transient_right_minimum:
		return (None, None)
	analseg = analsteps.truncate_end(segments.segment(max(analseg[0], sciseg[0] + config.transient_left_minimum), min(analseg[1], sciseg[1] - config.transient_right_minimum)))
	if not analseg.duration():
		return (None, None)

	# Construct the left and right transients as the minimum allowed
	# pieces before and after the analysis segment.
	ltransseg = segments.segment(analseg[0] - config.transient_left_minimum, analseg[0])
	rtransseg = segments.segment(analseg[1], analseg[1] + config.transient_right_minimum)

	# Define the data segment as the sum of the three, and check for
	# minimum length.
	dataseg = ltransseg + analseg + rtransseg
	if dataseg.duration() >= config.data_length_minimum:
		return (dataseg, analseg)

	# Attempt to satisfy the required minimum data length by increasing
	# the left transient.
	ltransseg = ltranssteps.add_start(ltransseg, min(config.data_length_minimum - dataseg.duration(), ltransseg[0] - sciseg[0]))
	dataseg = ltransseg + analseg + rtransseg
	if dataseg.duration() >= config.data_length_minimum:
		return (dataseg, analseg)

	# Attempt to satisfy the required minimum data length by increasing
	# the right transient.
	rtransseg = rtranssteps.add_end(rtransseg, min(config.data_length_minimum - dataseg.duration(), sciseg[1] - rtransseg[1]))
	dataseg = ltransseg + analseg + rtransseg
	if dataseg.duration() >= config.data_length_minimum:
		return (dataseg, analseg)

	return (None, None)


#
# Default DAG generator interface.
#

class DAGGeneratorConfig(onasys.DAGGenerator.DAGGeneratorConfig):
	def defaults(self):
		return [
			("data_dir", os.getcwd()),
			("extra_args", ""),
		]

	def parse(self):
		for key in ["data_dir", "extra_args", "generator"]:
			self.__dict__[key] = self.parser.get(self.section, key)
		for key in ["cusion_timeout", "data_length_minimum", "analysis_length_minimum", "analysis_length_increment", "transient_left_minimum", "transient_left_increment", "transient_right_minimum", "transient_right_increment"]:
			self.__dict__[key] = self.parser.getfloat(self.section, key)

	def validate(self):
		if self.data_length_minimum < self.analysis_length_minimum + self.transient_left_minimum + self.transient_right_minimum:
			raise ConfigError, ConfigErrorMsg(self, "sum of analysis_length, transient_left and transient_right minima excedes data_length_minimum")

class DAGGeneratorParameters(onasys.DAGGenerator.DAGGeneratorParameters):
	def __init__(self, segpair):
		self.dataseg, self.analseg = segpair

	def __repr__(self):
		return repr(self.dataseg)

def DAGGeneratorParameterList(clock, config, available, used):
	# Hack out the segmentlist we care about.
	available = available.get("available", segments.segmentlist())

	# Leave behind a cushion.
	if len(available):
		if float(clock.now() - available[-1][1]) < config.cusion_timeout:
			available -= segments.segmentlist([segments.segment(available[-1][1] - config.analysis_length_minimum - config.transient_right_minimum, available[-1][1])])

	# Calculate the unused analysis segments.
	analsegs = available - used.get("analysis", segments.segmentlist())

	# For each analysis segment, identify the science segment in which
	# it occurs.
	scisegs = segments.segmentlist(map(lambda seg: available[available.find(seg)], analsegs))

	# Construct a list of data segment and analysis segment pairs.
	pairs = map(lambda sciseg, analseg: compute_boundaries(config, sciseg, analseg), scisegs, analsegs)

	# Throw out null results.
	pairs = filter(lambda pair: pair != (None, None), pairs)

	# Return a list of DAGGeneratorParameters from the data and
	# analysis segments.
	return [DAGGeneratorParameters(pair) for pair in pairs]

class DAGGenerator(onasys.DAGGenerator.DAGGenerator):
	def dag_name(self):
		return self.tag + "-" + str(int(self.params.dataseg[0])) + "-" + str(int(self.params.dataseg.duration())) + ".dag"

	def generate_command(self):
		command = self.config.data_dir + "/" + self.config.generator + " -s " + str(self.params.dataseg[0]) + " -e " + str(self.params.dataseg[1]) + " -a " + str(self.params.analseg[0]) + " -b " + str(self.params.analseg[1]) + " -f " + self.dag_name() + " -t " + self.config.data_dir
		if self.config.extra_args:
			command += " " + self.config.extra_args
		return command

	def generate(self):
		logging.info(self.generate_command())
		child = popen2.Popen3(self.generate_command(), True)
		for line in child.childerr:
			logging.warning(line.rstrip())
		newlyused = segments.segmentlist()
		try:
			for line in child.fromchild:
				newlyused |= segments.segmentlist([segments.segment(*map(lal.LIGOTimeGPS, line.split()))])
		except Exception, e:
			raise onasys.DAGGenerator.DAGGeneratorError, "error parsing generator output \"" + line.rstrip() + "\": " + str(e)
		if len(newlyused) != 1:
			logging.warning("generator returned more than one analysis segment for removal")
		if newlyused.extent() not in self.params.analseg:
			logging.warning("analysis segment to remove exceeds bounds of " + str(self.params.analseg))
		status = child.wait()
		if not os.WIFEXITED(status) or os.WEXITSTATUS(status):
			raise onasys.DAGGenerator.DAGGeneratorError, "generator returned " + str(os.WEXITSTATUS(status))
		return {"analysis": newlyused}
